// Package postgres provides repository implementation for postgresql database
package postgres

import (
	"context"
	"fmt"

	"gitlab.mashroom.com/mashroom/backend/contract/internal/model"

	"github.com/golang-migrate/migrate/v4"
	"github.com/golang-migrate/migrate/v4/database/postgres"
	bindata "github.com/golang-migrate/migrate/v4/source/go_bindata"
	"github.com/jmoiron/sqlx"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/contract/internal/model/postgres/migrations"
)

type repositoryImpl struct {
	DB                   *sqlx.DB
	stmtContractGet      *sqlx.NamedStmt
	stmtContractCreate   *sqlx.NamedStmt
	stmtContractUpdate   *sqlx.NamedStmt
	stmtContractDelete   *sqlx.NamedStmt
	stmtContractRevision *sqlx.NamedStmt
	stmtContractList     *sqlx.NamedStmt
	stmtRevisionList     *sqlx.NamedStmt
	stmtRevisionLoad     *sqlx.NamedStmt
}

func prepareNamed(db *sqlx.DB, s string) (r *sqlx.NamedStmt, err error) {
	r, err = db.PrepareNamed(s)

	if err != nil {
		fmt.Printf("error %v preparing %s\n", err, s)
	}

	return
}

// NewRepository returns new postgresql repository
func NewRepository(db *sqlx.DB) (repos model.Repository, err error) {
	repo := &repositoryImpl{
		DB: db,
	}

	err = repo.Migrate()
	if err != nil {
		return nil, err
	}

	repo.stmtContractGet, err = prepareNamed(db, queryContractGet)
	if err != nil {
		return
	}

	repo.stmtContractCreate, err = prepareNamed(db, queryContractCreate)
	if err != nil {
		return
	}

	repo.stmtContractUpdate, err = prepareNamed(db, queryContractUpdate)
	if err != nil {
		return
	}

	repo.stmtContractDelete, err = prepareNamed(db, queryContractDelete)
	if err != nil {
		return
	}

	repo.stmtContractRevision, err = prepareNamed(db, queryContractRevision)
	if err != nil {
		return
	}

	repo.stmtContractList, err = prepareNamed(db, queryContractList)
	if err != nil {
		return
	}

	repo.stmtRevisionList, err = prepareNamed(db, queryRevisionList)
	if err != nil {
		return
	}

	repo.stmtRevisionLoad, err = prepareNamed(db, queryRevisionLoad)
	if err != nil {
		return
	}

	return repo, nil
}

func (repo *repositoryImpl) Migrate() error {
	source, err := bindata.WithInstance(bindata.Resource(migrations.AssetNames(), migrations.Asset))
	if err != nil {
		panic(err)
	}

	driver, err := postgres.WithInstance(repo.DB.DB, &postgres.Config{})
	if err != nil {
		panic(err)
	}

	m, err := migrate.NewWithInstance("go-bindata", source, "postgres", driver)
	if err != nil {
		panic(err)
	}

	err = m.Up()

	switch {
	case err == nil:
		log.Println("Migrated.")
	case err == migrate.ErrNoChange:
		log.Println("Migration not required.")
	default:
		return err
	}

	return nil
}

const queryContractGet = `
select
  contracts.id,
  contracts.parent,
  contracts.created_at,
  contracts.updated_at,
  contracts.name,
  contracts.description,
  contracts.last_revision,
  contracts.deleted
from contracts
where id = :id
`

func (repo *repositoryImpl) ContractGet(ctx context.Context, id uint64) (m *model.Contract, err error) {
	m = &model.Contract{
		ID: id,
	}

	err = repo.stmtContractGet.GetContext(ctx, m, m)
	if err != nil {
		return nil, err
	}

	return
}

const queryContractCreate = `
insert into contracts (
  id,
  parent,
  created_at,
  updated_at,
  name,
  description,
  last_revision,
  deleted
) values (
  default,
  :parent,
  now(),
  now(),
  :name,
  :description,
  0,
  false
) returning
  id,
  parent,
  created_at,
  updated_at,
  name,
  description,
  last_revision,
  deleted
`

func (repo *repositoryImpl) ContractCreate(ctx context.Context, contract *model.Contract) (err error) {
	err = repo.stmtContractCreate.GetContext(ctx, contract, contract)
	if err != nil {
		return err
	}

	return
}

const queryContractUpdate = `
update contracts set
  parent = :parent,
  name = :name,
  description = :description,
  updated_at = now()
where
  id = :id
returning
  id,
  parent,
  created_at,
  updated_at,
  name,
  description,
  last_revision,
  deleted
`

func (repo *repositoryImpl) ContractUpdate(ctx context.Context, contract *model.Contract) (err error) {
	err = repo.stmtContractUpdate.GetContext(ctx, contract, contract)
	if err != nil {
		return err
	}

	return
}

const queryContractDelete = `
update contracts
  set deleted = true
where
  contracts.id = :id and
  ((select count(*) from contracts where parent = :id) = 0 or :deleted)
`

func (repo *repositoryImpl) ContractDelete(ctx context.Context, id uint64, force bool) (err error) {
	_, err = repo.stmtContractDelete.ExecContext(ctx, &model.Contract{
		ID:      id,
		Deleted: force,
	})
	if err != nil {
		return err
	}

	return
}

const queryContractRevision = `
with rows as (
  insert into revisions(
    id,
    contract_id,
    created_at,
    text
  ) values (
    default,
    :contract_id,
    now(),
    :text
  ) returning id
) update contracts set 
  last_revision = (select id from rows),
  updated_at = now()
where id = :contract_id
returning
  id,
  parent,
  created_at,
  updated_at,
  name,
  description,
  last_revision,
  deleted
`

func (repo *repositoryImpl) ContractRevision(
	ctx context.Context,
	contractID uint64,
	text string,
) (m *model.Contract, err error) {
	m = &model.Contract{}

	err = repo.stmtContractRevision.GetContext(ctx, m, &model.Revision{
		ContractID: contractID,
		Text:       text,
	})
	if err != nil {
		return nil, err
	}

	return
}

const queryContractList = `
select 
  id,
  parent,
  created_at,
  updated_at,
  name,
  description,
  last_revision,
  deleted
from contracts where
  parent = :parent and
  (contracts.deleted = false or contracts.deleted = :deleted)
order by name asc
`

func (repo *repositoryImpl) ContractList(
	ctx context.Context,
	parentID uint64,
	deleted bool,
) (ms []*model.Contract, err error) {
	err = repo.stmtContractList.SelectContext(ctx, &ms, &model.Contract{
		Parent:  parentID,
		Deleted: deleted,
	})
	if err != nil {
		return nil, err
	}

	return
}

const queryRevisionList = `
select 
  revisions.id,
  revisions.contract_id,
  revisions.created_at
from revisions where 
  contract_id = :contract_id 
order by revisions.created_at desc
`

func (repo *repositoryImpl) RevisionList(ctx context.Context, contractID uint64) (ms []*model.Revision, err error) {
	err = repo.stmtRevisionList.SelectContext(ctx, &ms, &model.Revision{
		ContractID: contractID,
	})
	if err != nil {
		return nil, err
	}

	return
}

const queryRevisionLoad = `
select 
  revisions.id,
  revisions.contract_id,
  revisions.created_at,
  revisions.text
from revisions where 
  id = :id
`

func (repo *repositoryImpl) RevisionLoad(ctx context.Context, id uint64) (m *model.Revision, err error) {
	m = &model.Revision{}

	err = repo.stmtRevisionLoad.GetContext(ctx, m, &model.Revision{
		ID: id,
	})
	if err != nil {
		return nil, err
	}

	return
}
