package service

import (
	"bytes"
	"context"
	"encoding/json"
	"fmt"
	"html/template"
	"io/ioutil"
	"mime"
	"mime/multipart"
	"net/http"
	"strconv"
	"time"

	contract "gitlab.mashroom.com/mashroom/backend/contract/api"
	"gitlab.mashroom.com/mashroom/backend/contract/internal/model"
	document "gitlab.mashroom.com/mashroom/backend/document/api"
)

const (
	extHTML = "html"
	extPDF  = "pdf"
)

func (srv *serverImpl) Create(ctx context.Context, req *contract.CreateRequest) (*contract.CreateResponse, error) {
	parentID, err := strconv.ParseUint(req.GetParentId(), 10, 64)
	if err != nil {
		return nil, err
	}

	parent, err := srv.Repo.ContractGet(ctx, parentID)
	if err != nil {
		return nil, err
	}

	m := &model.Contract{
		Parent:      parent.ID,
		Name:        req.GetName(),
		Description: req.GetDescription(),
	}

	err = srv.Repo.ContractCreate(ctx, m)
	if err != nil {
		return nil, err
	}

	return &contract.CreateResponse{
		Info:   srv.renderContract(m),
		Status: contract.ContractStatus_SUCCESS,
	}, nil
}

func (srv *serverImpl) Update(ctx context.Context, req *contract.UpdateRequest) (*contract.UpdateResponse, error) {
	id, err := strconv.ParseUint(req.GetId(), 10, 64)
	if err != nil {
		return nil, err
	}

	parentID, err := strconv.ParseUint(req.GetParentId(), 10, 64)
	if err != nil {
		return nil, err
	}

	m, err := srv.Repo.ContractGet(ctx, id)
	if err != nil {
		return nil, err
	}

	parent, err := srv.Repo.ContractGet(ctx, parentID)
	if err != nil {
		return nil, err
	}

	m.Parent = parent.ID
	m.Name = req.GetName()
	m.Description = req.GetDescription()

	err = srv.Repo.ContractUpdate(ctx, m)
	if err != nil {
		return nil, err
	}

	return &contract.UpdateResponse{
		Info:   srv.renderContract(m),
		Status: contract.ContractStatus_SUCCESS,
	}, nil
}

func (srv *serverImpl) Edit(ctx context.Context, req *contract.EditRequest) (*contract.EditResponse, error) {
	id, err := strconv.ParseUint(req.GetId(), 10, 64)
	if err != nil {
		return nil, err
	}

	m, err := srv.Repo.ContractRevision(ctx, id, req.GetText())
	if err != nil {
		return nil, err
	}

	return &contract.EditResponse{
		Info:   srv.renderContract(m),
		Status: contract.ContractStatus_SUCCESS,
	}, nil
}

func (srv *serverImpl) History(ctx context.Context, req *contract.HistoryRequest) (*contract.HistoryResponse, error) {
	id, err := strconv.ParseUint(req.GetId(), 10, 64)
	if err != nil {
		return nil, err
	}

	m, err := srv.Repo.ContractGet(ctx, id)
	if err != nil {
		return nil, err
	}

	rs, err := srv.Repo.RevisionList(ctx, id)
	if err != nil {
		return nil, err
	}

	var revisions []*contract.ContractRevision
	for _, r := range rs {
		revisions = append(revisions, srv.renderRevision(r))
	}

	return &contract.HistoryResponse{
		Info:      srv.renderContract(m),
		Status:    contract.ContractStatus_SUCCESS,
		Revisions: revisions,
	}, nil
}

func (srv *serverImpl) List(ctx context.Context, req *contract.ListRequest) (*contract.ListResponse, error) {
	parentID, err := strconv.ParseUint(req.GetParentId(), 10, 64)
	if err != nil {
		return nil, err
	}

	m, err := srv.Repo.ContractGet(ctx, parentID)
	if err != nil {
		return nil, err
	}

	ms, err := srv.Repo.ContractList(ctx, parentID, req.GetDeleted())
	if err != nil {
		return nil, err
	}

	var contracts []*contract.ContractInfo

	for _, c := range ms {
		if c.ID != 0 {
			contracts = append(contracts, srv.renderContract(c))
		}
	}

	return &contract.ListResponse{
		Info:     srv.renderContract(m),
		Status:   contract.ContractStatus_SUCCESS,
		Children: contracts,
	}, nil
}

func (srv *serverImpl) Delete(ctx context.Context, req *contract.DeleteRequest) (*contract.DeleteResponse, error) {
	id, err := strconv.ParseUint(req.GetId(), 10, 64)
	if err != nil {
		return nil, err
	}

	err = srv.Repo.ContractDelete(ctx, id, req.GetForce())
	if err != nil {
		return nil, err
	}

	return &contract.DeleteResponse{
		Status: contract.ContractStatus_SUCCESS,
	}, nil
}

func (srv *serverImpl) weaverPDFRender(ctx context.Context, content []byte) ([]byte, error) {
	body := &bytes.Buffer{}

	mw := multipart.NewWriter(body)

	w, err := mw.CreateFormFile("file", "input.html")
	if err != nil {
		return nil, err
	}

	_, err = w.Write(content)
	if err != nil {
		return nil, err
	}

	_ = mw.Close()

	req, err := http.NewRequestWithContext(ctx, http.MethodPost, srv.Config.WeaverURL, body)
	if err != nil {
		return nil, err
	}

	req.Header.Set("content-type", mw.FormDataContentType())

	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		return nil, err
	}

	defer func() {
		_ = resp.Body.Close()
	}()

	return ioutil.ReadAll(resp.Body)
}

func (srv *serverImpl) Render(ctx context.Context, req *contract.RenderRequest) (*contract.RenderResponse, error) {
	id, err := strconv.ParseUint(req.GetId(), 10, 64)
	if err != nil {
		return nil, err
	}

	m, err := srv.Repo.ContractGet(ctx, id)
	if err != nil {
		return nil, err
	}

	rs, err := srv.Repo.RevisionLoad(ctx, m.LastRevision)
	if err != nil {
		return nil, err
	}

	t, err := template.New(m.Name).Parse(rs.Text)
	if err != nil {
		return nil, err
	}

	values := make(map[string]interface{})

	err = json.Unmarshal(req.GetValuesJson(), &values)
	if err != nil {
		return nil, err
	}

	sb := &bytes.Buffer{}

	err = t.Execute(sb, values)
	if err != nil {
		return nil, err
	}

	var bs []byte

	var ext string

	switch req.GetFormat() {
	case contract.ContractFormat_HTML:
		bs = sb.Bytes()
		ext = extHTML
	case contract.ContractFormat_PDF:
		bs, err = srv.weaverPDFRender(ctx, sb.Bytes())
		if err != nil {
			return nil, err
		}

		ext = extPDF
	default:
		return nil, nil
	}

	now := time.Now()

	name := req.GetName()
	desc := req.GetDescription()

	if name == "" {
		name = fmt.Sprintf("contract-%04d-%02d-%02d.%s", now.Year(), now.Month(), now.Day(), ext)
	}

	if desc == "" {
		desc = fmt.Sprintf(
			"Contract as of %04d-%02d-%02d, rendered in %s",
			now.Year(),
			now.Month(),
			now.Day(),
			ext,
		)
	}

	rdr := bytes.NewReader(bs)

	file, err := document.UploadFile(ctx, srv.Connections.Document, name, mime.TypeByExtension("."+ext), desc, rdr)
	if err != nil {
		return nil, fmt.Errorf("upload failure: %w", err)
	}

	return &contract.RenderResponse{
		Fileid:  file.GetId(),
		Success: true,
	}, nil
}
