package service

import (
	"context"
	"errors"
	"net/http"
	"strconv"
	"testing"

	"github.com/stretchr/testify/require"

	contract "gitlab.mashroom.com/mashroom/backend/contract/api"
	"gitlab.mashroom.com/mashroom/backend/contract/internal/model"
)

func Test_serverImpl_Create(t *testing.T) {
	type args struct {
		ctx context.Context
		req *contract.CreateRequest
	}
	tests := []struct { // nolint:wsl
		name            string
		defineDeps      func(t *testing.T) *testDependencies
		args            args
		mockPreparation func(t *testing.T, testDeps *testDependencies, arg *args)
		want            *contract.CreateResponse
		wantErr         bool
	}{
		{
			name: "ok",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.CreateRequest{
					ParentId:    "1",
					Name:        "test",
					Description: "test description",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				parentID, err := strconv.ParseUint(arg.req.GetParentId(), 10, 64)
				require.NoError(t, err)

				returnedByContractGet := &model.Contract{
					ID:           1,
					LastRevision: 10,
					Name:         "test_contract",
				}

				testDeps.repo.EXPECT().ContractGet(arg.ctx, parentID).
					Return(returnedByContractGet, nil)

				createContractModel := &model.Contract{
					Parent:      returnedByContractGet.ID,
					Name:        arg.req.GetName(),
					Description: arg.req.GetDescription(),
				}

				testDeps.repo.EXPECT().ContractCreate(arg.ctx, createContractModel).
					Return(nil)
			},
			want: &contract.CreateResponse{
				Info: &contract.ContractInfo{
					ParentId:     "1",
					Id:           "0",
					TimeCreated:  -62135596800,
					TimeUpdated:  -62135596800,
					Name:         "test",
					Description:  "test description",
					LastRevision: "0",
					Deleted:      false,
				},
				Status: contract.ContractStatus_SUCCESS,
			},
			wantErr: false,
		},
		{
			name: "error - invalid parent id",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.CreateRequest{
					ParentId:    "t",
					Name:        "test",
					Description: "test description",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
			},
			want:    nil,
			wantErr: true,
		},
		{
			name: "error - get contract",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.CreateRequest{
					ParentId:    "1",
					Name:        "test",
					Description: "test description",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				parentID, err := strconv.ParseUint(arg.req.GetParentId(), 10, 64)
				require.NoError(t, err)

				testDeps.repo.EXPECT().ContractGet(arg.ctx, parentID).
					Return(nil, errors.New("get contract error"))
			},
			want:    nil,
			wantErr: true,
		},
		{
			name: "error - create contract",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.CreateRequest{
					ParentId:    "1",
					Name:        "test",
					Description: "test description",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				parentID, err := strconv.ParseUint(arg.req.GetParentId(), 10, 64)
				require.NoError(t, err)

				returnedByContractGet := &model.Contract{
					ID:           1,
					LastRevision: 10,
					Name:         "test_contract",
				}

				testDeps.repo.EXPECT().ContractGet(arg.ctx, parentID).
					Return(returnedByContractGet, nil)

				createContractModel := &model.Contract{
					Parent:      returnedByContractGet.ID,
					Name:        arg.req.GetName(),
					Description: arg.req.GetDescription(),
				}

				testDeps.repo.EXPECT().ContractCreate(arg.ctx, createContractModel).
					Return(errors.New("create contract error"))
			},
			want:    nil,
			wantErr: true,
		},
	}
	for _, tt := range tests { // nolint:wsl
		tt := tt
		t.Run(tt.name, func(t *testing.T) {
			testDeps := tt.defineDeps(t)
			defer func() {
				testDeps.ctrl.Finish()
				testDeps.serverMock.Close()
			}()

			tt.mockPreparation(t, testDeps, &tt.args)

			srv := testDeps.srv
			got, err := srv.Create(tt.args.ctx, tt.args.req)
			if (err != nil) != tt.wantErr {
				t.Errorf("Create() error = %v, wantErr %v", err, tt.wantErr)
				return
			}

			require.Equalf(t, got, tt.want, "Create() got = %v, want %v", got, tt.want)
		})
	}
}
