package service

import (
	"context"
	"errors"
	"net/http"
	"strconv"
	"testing"

	"github.com/stretchr/testify/require"

	contract "gitlab.mashroom.com/mashroom/backend/contract/api"
)

func Test_serverImpl_Delete(t *testing.T) {
	type args struct {
		ctx context.Context
		req *contract.DeleteRequest
	}
	tests := []struct { // nolint:wsl
		name            string
		defineDeps      func(t *testing.T) *testDependencies
		args            args
		mockPreparation func(t *testing.T, testDeps *testDependencies, arg *args)
		want            *contract.DeleteResponse
		wantErr         bool
	}{
		{
			name: "ok",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.DeleteRequest{
					Id:    "1",
					Force: true,
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				id, err := strconv.ParseUint(arg.req.GetId(), 10, 64)
				require.NoError(t, err)

				testDeps.repo.EXPECT().ContractDelete(arg.ctx, id, arg.req.GetForce()).
					Return(nil)
			},
			want: &contract.DeleteResponse{
				Status: contract.ContractStatus_SUCCESS,
			},
			wantErr: false,
		},
		{
			name: "error - invalid id",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.DeleteRequest{
					Id:    "t",
					Force: true,
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {},
			want:            nil,
			wantErr:         true,
		},
		{
			name: "error - delete contract",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.DeleteRequest{
					Id:    "1",
					Force: true,
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				id, err := strconv.ParseUint(arg.req.GetId(), 10, 64)
				require.NoError(t, err)

				testDeps.repo.EXPECT().ContractDelete(arg.ctx, id, arg.req.GetForce()).
					Return(errors.New("delete contract error"))
			},
			want:    nil,
			wantErr: true,
		},
	}
	for _, tt := range tests { // nolint:wsl
		tt := tt
		t.Run(tt.name, func(t *testing.T) {
			testDeps := tt.defineDeps(t)
			defer func() {
				testDeps.ctrl.Finish()
				testDeps.serverMock.Close()
			}()

			tt.mockPreparation(t, testDeps, &tt.args)

			srv := testDeps.srv
			got, err := srv.Delete(tt.args.ctx, tt.args.req)
			if (err != nil) != tt.wantErr {
				t.Errorf("Delete() error = %v, wantErr %v", err, tt.wantErr)
				return
			}

			require.Equalf(t, tt.want, got, "Delete() got = %v, want %v", got, tt.want)
		})
	}
}
