package service

import (
	"context"
	"errors"
	"net/http"
	"strconv"
	"testing"

	"github.com/stretchr/testify/require"

	contract "gitlab.mashroom.com/mashroom/backend/contract/api"
	"gitlab.mashroom.com/mashroom/backend/contract/internal/model"
)

func Test_serverImpl_Edit(t *testing.T) {
	type args struct {
		ctx context.Context
		req *contract.EditRequest
	}
	tests := []struct { // nolint:wsl
		name            string
		defineDeps      func(t *testing.T) *testDependencies
		args            args
		mockPreparation func(t *testing.T, testDeps *testDependencies, arg *args)
		want            *contract.EditResponse
		wantErr         bool
	}{
		{
			name: "ok",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.EditRequest{
					Id:   "1",
					Text: "test",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				id, err := strconv.ParseUint(arg.req.GetId(), 10, 64)
				require.NoError(t, err)

				returnedContract := &model.Contract{
					ID:           1,
					Parent:       2,
					LastRevision: 10,
					Name:         "test",
				}

				testDeps.repo.EXPECT().ContractRevision(arg.ctx, id, arg.req.GetText()).
					Return(returnedContract, nil)
			},
			want: &contract.EditResponse{
				Info: &contract.ContractInfo{
					ParentId:     "2",
					Id:           "1",
					TimeCreated:  -62135596800,
					TimeUpdated:  -62135596800,
					Name:         "test",
					Description:  "",
					LastRevision: "10",
					Deleted:      false,
				},
				Status: contract.ContractStatus_SUCCESS,
			},
			wantErr: false,
		},
		{
			name: "error - invalid id",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.EditRequest{
					Id:   "t",
					Text: "test",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {},
			want:            nil,
			wantErr:         true,
		},
		{
			name: "error - contract revision",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.EditRequest{
					Id:   "1",
					Text: "test",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				id, err := strconv.ParseUint(arg.req.GetId(), 10, 64)
				require.NoError(t, err)

				testDeps.repo.EXPECT().ContractRevision(arg.ctx, id, arg.req.GetText()).
					Return(nil, errors.New("contract revision error"))
			},
			want:    nil,
			wantErr: true,
		},
	}
	for _, tt := range tests { // nolint:wsl
		tt := tt
		t.Run(tt.name, func(t *testing.T) {
			testDeps := tt.defineDeps(t)
			defer func() {
				testDeps.ctrl.Finish()
				testDeps.serverMock.Close()
			}()

			tt.mockPreparation(t, testDeps, &tt.args)

			srv := testDeps.srv
			got, err := srv.Edit(tt.args.ctx, tt.args.req)
			if (err != nil) != tt.wantErr {
				t.Errorf("Edit() error = %v, wantErr %v", err, tt.wantErr)
				return
			}

			require.Equalf(t, tt.want, got, "Edit() got = %v, want %v", got, tt.want)
		})
	}
}
