package service

import (
	"context"
	"errors"
	"net/http"
	"strconv"
	"testing"

	"gitlab.mashroom.com/mashroom/backend/contract/internal/model"

	"github.com/stretchr/testify/require"

	contract "gitlab.mashroom.com/mashroom/backend/contract/api"
)

func Test_serverImpl_History(t *testing.T) {
	type args struct {
		ctx context.Context
		req *contract.HistoryRequest
	}
	tests := []struct { // nolint:wsl
		name            string
		defineDeps      func(t *testing.T) *testDependencies
		args            args
		mockPreparation func(t *testing.T, testDeps *testDependencies, arg *args)
		want            *contract.HistoryResponse
		wantErr         bool
	}{
		{
			name: "ok",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.HistoryRequest{
					Id:    "1",
					Depth: 2,
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				id, err := strconv.ParseUint(arg.req.GetId(), 10, 64)
				require.NoError(t, err)

				returnedContract := &model.Contract{
					ID: 1,
				}
				testDeps.repo.EXPECT().ContractGet(arg.ctx, id).
					Return(returnedContract, nil)

				var revisions = []*model.Revision{
					{
						ID:   1,
						Text: "test1",
					},
					{
						ID:   2,
						Text: "test2",
					},
				}

				testDeps.repo.EXPECT().RevisionList(arg.ctx, id).
					Return(revisions, nil)
			},
			want: &contract.HistoryResponse{
				Info: &contract.ContractInfo{
					ParentId:     "0",
					Id:           "1",
					TimeCreated:  -62135596800,
					TimeUpdated:  -62135596800,
					Name:         "",
					Description:  "",
					LastRevision: "0",
					Deleted:      false,
				},
				Revisions: []*contract.ContractRevision{
					{
						RevisionId:    "1",
						TimeCommitted: -62135596800,
						Text:          "test1",
					},
					{
						RevisionId:    "2",
						TimeCommitted: -62135596800,
						Text:          "test2",
					},
				},
				Status: contract.ContractStatus_SUCCESS,
			},
			wantErr: false,
		},
		{
			name: "error - invalid id",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.HistoryRequest{
					Id:    "t",
					Depth: 2,
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {},
			want:            nil,
			wantErr:         true,
		},
		{
			name: "error - get contract",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.HistoryRequest{
					Id:    "1",
					Depth: 2,
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				id, err := strconv.ParseUint(arg.req.GetId(), 10, 64)
				require.NoError(t, err)

				testDeps.repo.EXPECT().ContractGet(arg.ctx, id).
					Return(nil, errors.New("get contract error"))
			},
			want:    nil,
			wantErr: true,
		},
		{
			name: "error - revision list",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.HistoryRequest{
					Id:    "1",
					Depth: 2,
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				id, err := strconv.ParseUint(arg.req.GetId(), 10, 64)
				require.NoError(t, err)

				returnedContract := &model.Contract{
					ID: 1,
				}
				testDeps.repo.EXPECT().ContractGet(arg.ctx, id).
					Return(returnedContract, nil)

				testDeps.repo.EXPECT().RevisionList(arg.ctx, id).
					Return(nil, errors.New("revision list error"))
			},
			want:    nil,
			wantErr: true,
		},
	}
	for _, tt := range tests { // nolint:wsl
		tt := tt
		t.Run(tt.name, func(t *testing.T) {
			testDeps := tt.defineDeps(t)
			defer func() {
				testDeps.ctrl.Finish()
				testDeps.serverMock.Close()
			}()

			tt.mockPreparation(t, testDeps, &tt.args)

			srv := testDeps.srv
			got, err := srv.History(tt.args.ctx, tt.args.req)
			if (err != nil) != tt.wantErr {
				t.Errorf("History() error = %v, wantErr %v", err, tt.wantErr)
				return
			}

			require.Equalf(t, tt.want, got, "History() got = %v, want %v", got, tt.want)
		})
	}
}
