package service

import (
	"context"
	"errors"
	"net/http"
	"strconv"
	"testing"

	"github.com/stretchr/testify/require"

	contract "gitlab.mashroom.com/mashroom/backend/contract/api"
	"gitlab.mashroom.com/mashroom/backend/contract/internal/model"
)

func Test_serverImpl_List(t *testing.T) {
	type args struct {
		ctx context.Context
		req *contract.ListRequest
	}
	tests := []struct { // nolint:wsl
		name            string
		defineDeps      func(t *testing.T) *testDependencies
		args            args
		mockPreparation func(t *testing.T, testDeps *testDependencies, arg *args)
		want            *contract.ListResponse
		wantErr         bool
	}{
		{
			name: "ok",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.ListRequest{
					ParentId: "1",
					Deleted:  true,
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				parentID, err := strconv.ParseUint(arg.req.GetParentId(), 10, 64)
				require.NoError(t, err)

				returnedContract := &model.Contract{
					ID: parentID,
				}
				testDeps.repo.EXPECT().ContractGet(arg.ctx, parentID).
					Return(returnedContract, nil)

				list := []*model.Contract{
					{ID: 2, Deleted: true}, {ID: 3, Deleted: true},
				}
				testDeps.repo.EXPECT().ContractList(arg.ctx, parentID, arg.req.GetDeleted()).
					Return(list, nil)
			},
			want: &contract.ListResponse{
				Info: &contract.ContractInfo{
					ParentId:     "0",
					Id:           "1",
					TimeCreated:  -62135596800,
					TimeUpdated:  -62135596800,
					LastRevision: "0",
					Deleted:      false,
				},
				Status: contract.ContractStatus_SUCCESS,
				Children: []*contract.ContractInfo{
					{
						ParentId:     "0",
						Id:           "2",
						TimeCreated:  -62135596800,
						TimeUpdated:  -62135596800,
						LastRevision: "0",
						Deleted:      true,
					},
					{
						ParentId:     "0",
						Id:           "3",
						TimeCreated:  -62135596800,
						TimeUpdated:  -62135596800,
						LastRevision: "0",
						Deleted:      true,
					},
				},
			},
			wantErr: false,
		},
		{
			name: "error - invalid id",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.ListRequest{
					ParentId: "t",
					Deleted:  true,
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
			},
			want:    nil,
			wantErr: true,
		},
		{
			name: "error - get contract",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.ListRequest{
					ParentId: "1",
					Deleted:  true,
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				parentID, err := strconv.ParseUint(arg.req.GetParentId(), 10, 64)
				require.NoError(t, err)

				testDeps.repo.EXPECT().ContractGet(arg.ctx, parentID).
					Return(nil, errors.New("get contract error"))
			},
			want:    nil,
			wantErr: true,
		},
		{
			name: "error - contract list",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.ListRequest{
					ParentId: "1",
					Deleted:  true,
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				parentID, err := strconv.ParseUint(arg.req.GetParentId(), 10, 64)
				require.NoError(t, err)

				returnedContract := &model.Contract{
					ID: parentID,
				}
				testDeps.repo.EXPECT().ContractGet(arg.ctx, parentID).
					Return(returnedContract, nil)

				testDeps.repo.EXPECT().ContractList(arg.ctx, parentID, arg.req.GetDeleted()).
					Return(nil, errors.New("contract list error"))
			},
			want:    nil,
			wantErr: true,
		},
	}
	for _, tt := range tests { // nolint:wsl
		tt := tt
		t.Run(tt.name, func(t *testing.T) {
			testDeps := tt.defineDeps(t)
			defer func() {
				testDeps.ctrl.Finish()
				testDeps.serverMock.Close()
			}()

			tt.mockPreparation(t, testDeps, &tt.args)

			srv := testDeps.srv
			got, err := srv.List(tt.args.ctx, tt.args.req)
			if (err != nil) != tt.wantErr {
				t.Errorf("List() error = %v, wantErr %v", err, tt.wantErr)
				return
			}

			require.Equalf(t, tt.want, got, "List() got = %v, want %v", got, tt.want)
		})
	}
}
