package service

import (
	"context"
	"errors"
	"net/http"
	"strconv"
	"testing"
	"time"

	"github.com/golang/mock/gomock"
	"gitlab.mashroom.com/mashroom/backend/contract/internal/model"
	document "gitlab.mashroom.com/mashroom/backend/document/api"

	"github.com/stretchr/testify/require"

	contract "gitlab.mashroom.com/mashroom/backend/contract/api"
)

const tpl = `
<!DOCTYPE html>
<html>
	<head>
		<meta charset="UTF-8">
		<title>{.Title}}</title>
	</head>
	<body>
		{range .Items}}<div>{{ . }}</div>{{else}}<div><strong>no rows</strong></div>{{end}}
	</body>
</html>`

func Test_serverImpl_Render(t *testing.T) {
	type args struct {
		ctx context.Context
		req *contract.RenderRequest
	}
	tests := []struct { // nolint:wsl
		name            string
		defineDeps      func(t *testing.T) *testDependencies
		args            args
		mockPreparation func(t *testing.T, testDeps *testDependencies, arg *args)
		want            *contract.RenderResponse
		wantErr         bool
	}{
		{
			name: "ok for pdf",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {
					_, _ = w.Write([]byte(`{"testk":"testv"}`))
				}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.RenderRequest{
					Id:          "10",
					Format:      contract.ContractFormat_PDF,
					ValuesJson:  []byte(`{"test":"value"}`),
					Name:        "test",
					Description: "test description",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				id, err := strconv.ParseUint(arg.req.GetId(), 10, 64)
				require.NoError(t, err)

				returnedByContractGet := &model.Contract{
					ID:           1,
					LastRevision: 10,
					Name:         "test_contract",
				}

				testDeps.repo.EXPECT().ContractGet(arg.ctx, id).Return(returnedByContractGet, nil)

				returnedByRevisionLoad := &model.Revision{
					ID:         1,
					Text:       "test",
					ContractID: 1,
					CreatedAt:  time.Now(),
				}
				testDeps.repo.EXPECT().RevisionLoad(arg.ctx, returnedByContractGet.LastRevision).
					Return(returnedByRevisionLoad, nil)

				testDeps.docCli.EXPECT().FileUpload(arg.ctx).
					Return(testDeps.docUploadCli, nil)

				testDeps.docUploadCli.EXPECT().Send(gomock.Any()).Return(nil).AnyTimes()

				testDeps.docUploadCli.EXPECT().CloseAndRecv().
					Return(&document.FileUploadResponse{
						Success: true,
						File:    &document.File{Id: "1", Name: "test"},
					}, nil)
			},
			want: &contract.RenderResponse{
				Fileid:  "1",
				Success: true,
			},
			wantErr: false,
		},
		{
			name: "ok for html",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {
					_, _ = w.Write([]byte(`{"testk":"testv"}`))
				}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.RenderRequest{
					Id:          "10",
					Format:      contract.ContractFormat_HTML,
					ValuesJson:  []byte(`{"test":"value"}`),
					Name:        "test",
					Description: "test description",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				id, err := strconv.ParseUint(arg.req.GetId(), 10, 64)
				require.NoError(t, err)

				returnedByContractGet := &model.Contract{
					ID:           1,
					LastRevision: 10,
					Name:         "test_contract",
				}

				testDeps.repo.EXPECT().ContractGet(arg.ctx, id).Return(returnedByContractGet, nil)

				returnedByRevisionLoad := &model.Revision{
					ID:         1,
					Text:       "test",
					ContractID: 1,
					CreatedAt:  time.Now(),
				}
				testDeps.repo.EXPECT().RevisionLoad(arg.ctx, returnedByContractGet.LastRevision).
					Return(returnedByRevisionLoad, nil)

				testDeps.docCli.EXPECT().FileUpload(arg.ctx).
					Return(testDeps.docUploadCli, nil)

				testDeps.docUploadCli.EXPECT().Send(gomock.Any()).Return(nil).AnyTimes()

				testDeps.docUploadCli.EXPECT().CloseAndRecv().
					Return(&document.FileUploadResponse{
						Success: true,
						File:    &document.File{Id: "1", Name: "test"},
					}, nil)
			},
			want: &contract.RenderResponse{
				Fileid:  "1",
				Success: true,
			},
			wantErr: false,
		},
		{
			name: "error invalid request id parameter",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {
					_, _ = w.Write([]byte(`{"testk":"testv"}`))
				}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.RenderRequest{
					Id:          "ttt",
					Format:      contract.ContractFormat_HTML,
					ValuesJson:  []byte(`{"test":"value"}`),
					Name:        "test",
					Description: "test description",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {},
			want:            nil,
			wantErr:         true,
		},
		{
			name: "error contract get",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {
					_, _ = w.Write([]byte(`{"testk":"testv"}`))
				}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.RenderRequest{
					Id:          "10",
					Format:      contract.ContractFormat_PDF,
					ValuesJson:  []byte(`{"test":"value"}`),
					Name:        "test",
					Description: "test description",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				id, err := strconv.ParseUint(arg.req.GetId(), 10, 64)
				require.NoError(t, err)

				testDeps.repo.EXPECT().ContractGet(arg.ctx, id).Return(nil, errors.New("contract get error"))
			},
			want:    nil,
			wantErr: true,
		},
		{
			name: "error revision load",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {
					_, _ = w.Write([]byte(`{"testk":"testv"}`))
				}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.RenderRequest{
					Id:          "10",
					Format:      contract.ContractFormat_PDF,
					ValuesJson:  []byte(`{"test":"value"}`),
					Name:        "test",
					Description: "test description",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				id, err := strconv.ParseUint(arg.req.GetId(), 10, 64)
				require.NoError(t, err)

				returnedByContractGet := &model.Contract{
					ID:           1,
					LastRevision: 10,
					Name:         "test_contract",
				}

				testDeps.repo.EXPECT().ContractGet(arg.ctx, id).Return(returnedByContractGet, nil)

				testDeps.repo.EXPECT().RevisionLoad(arg.ctx, returnedByContractGet.LastRevision).
					Return(nil, errors.New("revision load error"))
			},
			want:    nil,
			wantErr: true,
		},
		{
			name: "error template parse",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {
					_, _ = w.Write([]byte(`{"testk":"testv"}`))
				}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.RenderRequest{
					Id:          "10",
					Format:      contract.ContractFormat_PDF,
					ValuesJson:  []byte(`{"test":"value"}`),
					Name:        "test",
					Description: "test description",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				id, err := strconv.ParseUint(arg.req.GetId(), 10, 64)
				require.NoError(t, err)

				returnedByContractGet := &model.Contract{
					ID:           1,
					LastRevision: 10,
					Name:         "test_contract",
				}

				testDeps.repo.EXPECT().ContractGet(arg.ctx, id).Return(returnedByContractGet, nil)

				returnedByRevisionLoad := &model.Revision{
					ID:         1,
					Text:       tpl,
					ContractID: 1,
					CreatedAt:  time.Now(),
				}
				testDeps.repo.EXPECT().RevisionLoad(arg.ctx, returnedByContractGet.LastRevision).
					Return(returnedByRevisionLoad, nil)
			},
			want:    nil,
			wantErr: true,
		},
		{
			name: "error - render request json values",
			defineDeps: func(t *testing.T) *testDependencies {
				weaverHandler := func(w http.ResponseWriter, r *http.Request) {
					_, _ = w.Write([]byte(`{"testk":"testv"}`))
				}

				return newDependencies(t, weaverHandler)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.RenderRequest{
					Id:          "10",
					Format:      contract.ContractFormat_PDF,
					ValuesJson:  []byte(`{"test":nil}`),
					Name:        "test",
					Description: "test description",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				id, err := strconv.ParseUint(arg.req.GetId(), 10, 64)
				require.NoError(t, err)

				returnedByContractGet := &model.Contract{
					ID:           1,
					LastRevision: 10,
					Name:         "test_contract",
				}

				testDeps.repo.EXPECT().ContractGet(arg.ctx, id).Return(returnedByContractGet, nil)

				returnedByRevisionLoad := &model.Revision{
					ID:         1,
					Text:       "test",
					ContractID: 1,
					CreatedAt:  time.Now(),
				}
				testDeps.repo.EXPECT().RevisionLoad(arg.ctx, returnedByContractGet.LastRevision).
					Return(returnedByRevisionLoad, nil)
			},
			want:    nil,
			wantErr: true,
		},
	}
	for _, tt := range tests { // nolint:wsl
		tt := tt
		t.Run(tt.name, func(t *testing.T) {
			testDeps := tt.defineDeps(t)
			defer func() {
				testDeps.ctrl.Finish()
				testDeps.serverMock.Close()
			}()

			tt.mockPreparation(t, testDeps, &tt.args)

			srv := testDeps.srv
			got, err := srv.Render(tt.args.ctx, tt.args.req)
			if (err != nil) != tt.wantErr {
				t.Errorf("Render() error = %v, wantErr %v", err, tt.wantErr)
				return
			}

			require.Equalf(t, got, tt.want, "Render() got = %v, want %v", got, tt.want)
		})
	}
}
