package service

import (
	"context"
	"errors"
	"net/http"
	"strconv"
	"testing"

	"github.com/stretchr/testify/require"

	contract "gitlab.mashroom.com/mashroom/backend/contract/api"
	"gitlab.mashroom.com/mashroom/backend/contract/internal/model"
)

func Test_serverImpl_Update(t *testing.T) {
	type args struct {
		ctx context.Context
		req *contract.UpdateRequest
	}
	tests := []struct { // nolint:wsl
		name            string
		defineDeps      func(t *testing.T) *testDependencies
		args            args
		mockPreparation func(t *testing.T, testDeps *testDependencies, arg *args)
		want            *contract.UpdateResponse
		wantErr         bool
	}{
		{
			name: "ok",
			defineDeps: func(t *testing.T) *testDependencies {
				h := func(w http.ResponseWriter, r *http.Request) {}
				return newDependencies(t, h)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.UpdateRequest{
					Id:          "1",
					ParentId:    "2",
					Name:        "test",
					Description: "test description",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				id, err := strconv.ParseUint(arg.req.GetId(), 10, 64)
				require.NoError(t, err)

				returnedByContractGet := &model.Contract{
					ID:           1,
					LastRevision: 10,
					Name:         "test_contract",
				}

				testDeps.repo.EXPECT().ContractGet(arg.ctx, id).
					Return(returnedByContractGet, nil)

				parentID, err := strconv.ParseUint(arg.req.GetParentId(), 10, 64)
				require.NoError(t, err)

				returnedByContractGetForParent := &model.Contract{
					ID:           2,
					LastRevision: 11,
					Name:         "test_contract2",
				}

				testDeps.repo.EXPECT().ContractGet(arg.ctx, parentID).
					Return(returnedByContractGetForParent, nil)

				returnedByContractGet.Parent = returnedByContractGetForParent.ID
				returnedByContractGet.Name = arg.req.Name
				returnedByContractGet.Description = arg.req.Description

				testDeps.repo.EXPECT().ContractUpdate(arg.ctx, returnedByContractGet).Return(nil)
			},
			want: &contract.UpdateResponse{
				Info: &contract.ContractInfo{
					ParentId:     "2",
					Id:           "1",
					TimeCreated:  -62135596800,
					TimeUpdated:  -62135596800,
					Name:         "test",
					Description:  "test description",
					LastRevision: "10",
					Deleted:      false,
				},
				Status: contract.ContractStatus_SUCCESS,
			},
			wantErr: false,
		},
		{
			name: "error - invalid id",
			defineDeps: func(t *testing.T) *testDependencies {
				h := func(w http.ResponseWriter, r *http.Request) {}
				return newDependencies(t, h)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.UpdateRequest{
					Id:          "t",
					ParentId:    "2",
					Name:        "test",
					Description: "test description",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {},
			want:            nil,
			wantErr:         true,
		},
		{
			name: "error - invalid parent id",
			defineDeps: func(t *testing.T) *testDependencies {
				h := func(w http.ResponseWriter, r *http.Request) {}
				return newDependencies(t, h)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.UpdateRequest{
					Id:          "1",
					ParentId:    "t",
					Name:        "test",
					Description: "test description",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {},
			want:            nil,
			wantErr:         true,
		},
		{
			name: "error - get contract",
			defineDeps: func(t *testing.T) *testDependencies {
				h := func(w http.ResponseWriter, r *http.Request) {}
				return newDependencies(t, h)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.UpdateRequest{
					Id:          "1",
					ParentId:    "2",
					Name:        "test",
					Description: "test description",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				id, err := strconv.ParseUint(arg.req.GetId(), 10, 64)
				require.NoError(t, err)

				testDeps.repo.EXPECT().ContractGet(arg.ctx, id).
					Return(nil, errors.New("get contract error"))
			},
			want:    nil,
			wantErr: true,
		},
		{
			name: "error - get parent contract",
			defineDeps: func(t *testing.T) *testDependencies {
				h := func(w http.ResponseWriter, r *http.Request) {}
				return newDependencies(t, h)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.UpdateRequest{
					Id:          "1",
					ParentId:    "2",
					Name:        "test",
					Description: "test description",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				id, err := strconv.ParseUint(arg.req.GetId(), 10, 64)
				require.NoError(t, err)

				returnedByContractGet := &model.Contract{
					ID:           1,
					LastRevision: 10,
					Name:         "test_contract",
				}

				testDeps.repo.EXPECT().ContractGet(arg.ctx, id).
					Return(returnedByContractGet, nil)

				parentID, err := strconv.ParseUint(arg.req.GetParentId(), 10, 64)
				require.NoError(t, err)

				testDeps.repo.EXPECT().ContractGet(arg.ctx, parentID).
					Return(nil, errors.New("get parent contract error"))
			},
			want:    nil,
			wantErr: true,
		},
		{
			name: "error - update contract",
			defineDeps: func(t *testing.T) *testDependencies {
				h := func(w http.ResponseWriter, r *http.Request) {}
				return newDependencies(t, h)
			},
			args: args{
				ctx: context.Background(),
				req: &contract.UpdateRequest{
					Id:          "1",
					ParentId:    "2",
					Name:        "test",
					Description: "test description",
				},
			},
			mockPreparation: func(t *testing.T, testDeps *testDependencies, arg *args) {
				id, err := strconv.ParseUint(arg.req.GetId(), 10, 64)
				require.NoError(t, err)

				returnedByContractGet := &model.Contract{
					ID:           1,
					LastRevision: 10,
					Name:         "test_contract",
				}

				testDeps.repo.EXPECT().ContractGet(arg.ctx, id).
					Return(returnedByContractGet, nil)

				parentID, err := strconv.ParseUint(arg.req.GetParentId(), 10, 64)
				require.NoError(t, err)

				returnedByContractGetForParent := &model.Contract{
					ID:           2,
					LastRevision: 11,
					Name:         "test_contract2",
				}

				testDeps.repo.EXPECT().ContractGet(arg.ctx, parentID).
					Return(returnedByContractGetForParent, nil)

				returnedByContractGet.Parent = returnedByContractGetForParent.ID
				returnedByContractGet.Name = arg.req.Name
				returnedByContractGet.Description = arg.req.Description

				testDeps.repo.EXPECT().ContractUpdate(arg.ctx, returnedByContractGet).
					Return(errors.New("update contract error"))
			},
			want:    nil,
			wantErr: true,
		},
	}
	for _, tt := range tests { // nolint:wsl
		tt := tt
		t.Run(tt.name, func(t *testing.T) {
			testDeps := tt.defineDeps(t)
			defer func() {
				testDeps.ctrl.Finish()
				testDeps.serverMock.Close()
			}()

			tt.mockPreparation(t, testDeps, &tt.args)

			srv := testDeps.srv
			got, err := srv.Update(tt.args.ctx, tt.args.req)
			if (err != nil) != tt.wantErr {
				t.Errorf("Update() error = %v, wantErr %v", err, tt.wantErr)
				return
			}

			require.Equalf(t, tt.want, got, "Update() got = %v, want %v", got, tt.want)
		})
	}
}
